
static char help[] = "Computes the second smallest eigenvalue of the Laplacian of a graph.\n\n";
  
#include <petscmat.h>   

#include <petscis.h> 

#include <petscsys.h>

#include <slepceps.h>

#include <stdio.h>

#include <stdlib.h>   

/*
    Computes elementwise absolute value of Matrix
 */

PetscErrorCode absmat(Mat A,Mat *AbsA)                                              
{
  PetscErrorCode ierr;
  PetscInt       n,i,nc,j,rstart,rend;
  const PetscInt *aj;                                                                
  const PetscScalar *aa;
  PetscScalar *absaa; 
  
  ierr = MatGetSize(A,&n,NULL);CHKERRQ(ierr);
  ierr = MatDuplicate(A,MAT_COPY_VALUES,AbsA);CHKERRQ(ierr);
       
  
  /* copy over the matrix entries from the matrix A to matrix AbsA*/
  ierr = MatGetOwnershipRange(A,&rstart,&rend);CHKERRQ(ierr);                  
  

  for (i=rstart; i<rend; i++) {                                                 
    ierr = MatGetRow(A,i,&nc,&aj,&aa);CHKERRQ(ierr);                               
    
    /* replace the nonzero values with their absolute values*/
    
    PetscMalloc1(nc,&absaa);
    for (j=0; j<nc; j++){
        absaa[j] = fabs(aa[j]);
    }
    
    ierr = MatSetValues(*AbsA,1,&i,nc,aj,absaa,INSERT_VALUES);CHKERRQ(ierr);    
    ierr = MatRestoreRow(A,i,&nc,&aj,&aa);CHKERRQ(ierr);
  }
                    
  return(0);          
}



int main (int argc,char **argv)                                                     
{
  EPS               eps;                             
  Mat               A=NULL,Atr,SymmA,Atrabs,Aabs,L,PL,NSymmA,NSymmAtrabs,AL;    
  int               s;                
  Vec               x,D,DD,vr,DDD;             
  EPSType           type;
  PetscInt          i,nev,*idx,mm,nn,rw,ncols,j,siz; 
  PetscScalar       kr,none=-1.0,*arr;                                          
  const PetscScalar *vals;
  const PetscInt    *cols;
  IS                is,partitioning;
  PetscViewer       fd; 
  char              file[PETSC_MAX_PATH_LEN];
  PetscBool         flg,isEqual,unw,wgh;  
  PetscErrorCode    ierr;
  MatPartitioning   part;


  ierr = SlepcInitialize(&argc,&argv,(char*)0,help);if (ierr) return ierr;
  
  // Read matrix from Petsc Binary File                                          

  ierr = PetscInitialize(&argc,&argv,(char*)0,help);if (ierr) return ierr;
  
  /*
     Determine files from which we read matrix
  */
  
  ierr = PetscOptionsGetString(NULL,NULL,"-f",file,PETSC_MAX_PATH_LEN,&flg);CHKERRQ(ierr);
  if (!flg) SETERRQ(PETSC_COMM_WORLD,1,"Must indicate binary file with the -f option");

  /*
     Open binary file.  Note that we use FILE_MODE_READ to indicate
     reading from this file.
  */
  
  ierr = PetscViewerBinaryOpen(PETSC_COMM_WORLD,file,FILE_MODE_READ,&fd);CHKERRQ(ierr);

  /*
    Load the matrix; then destroy the viewer.
  */
  
  ierr = MatCreate(PETSC_COMM_WORLD,&A);CHKERRQ(ierr);                               
  ierr = MatSetOptionsPrefix(A,"a_");CHKERRQ(ierr);
  ierr = MatSetFromOptions(A);CHKERRQ(ierr);
  ierr = MatLoad(A,fd);CHKERRQ(ierr);
  ierr = PetscViewerDestroy(&fd);CHKERRQ(ierr);
    
    /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
                Laplacian with Weighted and Unweighted Options
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
    
    /* Symmetry check */

    ierr = MatTranspose(A,MAT_INITIAL_MATRIX,&Atr);CHKERRQ(ierr);
    ierr = MatDuplicate(Atr,MAT_COPY_VALUES,&SymmA);CHKERRQ(ierr);
    flg = PETSC_TRUE;                                                           
    PetscOptionsGetBool(NULL,NULL, "-check_symmetry", &flg,NULL);
    if (flg) {
        ierr = MatEqual(A,Atr,&isEqual);CHKERRQ(ierr);
        if (isEqual) {
        ierr = PetscPrintf(PETSC_COMM_WORLD,"Input matrix is symmetric\n\n");CHKERRQ(ierr);
        ierr = absmat(A,&SymmA);CHKERRQ(ierr);

        } else {ierr = PetscPrintf(PETSC_COMM_WORLD,"Input matrix is not symmetric\n\n");CHKERRQ(ierr);
                ierr = absmat(Atr,&Atrabs);CHKERRQ(ierr);
                ierr = absmat(A,&Aabs);CHKERRQ(ierr);
                ierr = MatDuplicate(Atrabs,MAT_COPY_VALUES,&SymmA);CHKERRQ(ierr);
                ierr = MatAXPY(SymmA,1.,Aabs,DIFFERENT_NONZERO_PATTERN);CHKERRQ(ierr);  
        }

    }
    
    PetscOptionsHasName(NULL,NULL,"-unweighted",&unw);
    PetscOptionsHasName(NULL,NULL,"-weighted",&wgh);

    if (unw) { /*unweighted Laplacian*/

        /* DD = diag(SymmA);*/
        
        ierr = MatCreateVecs(SymmA,&DD,NULL);CHKERRQ(ierr);
        ierr = VecZeroEntries(DD);CHKERRQ(ierr);                                
        ierr = MatAssemblyBegin(SymmA,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
        ierr = MatAssemblyEnd(SymmA,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
        ierr = MatGetDiagonal(SymmA,DD);CHKERRQ(ierr);                          
        
        /* L = SymmN-diag(DD) */
    
        ierr = VecScale(DD,none);CHKERRQ(ierr);                                 
        ierr = MatDuplicate(SymmA,MAT_COPY_VALUES,&L);CHKERRQ(ierr);
        ierr = MatSetOption(L,MAT_NEW_NONZERO_ALLOCATION_ERR,PETSC_FALSE);CHKERRQ(ierr);
        ierr = MatDiagonalSet(L,DD,ADD_VALUES);CHKERRQ(ierr);
    
        /*[row,col,values] = find(L)
        values = -1;
        RestNew = sparse(row,col,values,x,y);*/

        ierr = MatGetOwnershipRange(L,&mm,&nn);CHKERRQ(ierr);                         
        
        ierr = MatDuplicate(L,MAT_DO_NOT_COPY_VALUES,&NSymmA);CHKERRQ(ierr);  
        
        for(rw = mm; rw<nn; ++rw){                                                      
            ierr = MatGetRow(L,rw,&ncols,&cols,&vals);CHKERRQ(ierr);                 
            s = ncols;
            ierr = PetscMalloc1(s,&arr);CHKERRQ(ierr);                              
            for(j=0;j<s;++j){                                                       
                arr[j]=-1.0;
            }
            ierr = MatSetOption(NSymmA,MAT_NEW_NONZERO_ALLOCATION_ERR,PETSC_FALSE);CHKERRQ(ierr); //ERROR FOR ODD NUMBERED SIZES
            ierr = MatSetValues(NSymmA,1,&rw,ncols,cols,arr,INSERT_VALUES);CHKERRQ(ierr);  
            ierr = MatRestoreRow(L,rw,&ncols,&cols,&vals);CHKERRQ(ierr);
        }

        ierr = MatAssemblyBegin(NSymmA,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);       //ERROR FOR ODD NUMBERED SIZES
        ierr = MatAssemblyEnd(NSymmA,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
        
        ierr = MatCreateVecs(NSymmA,&DDD,NULL);CHKERRQ(ierr);
        ierr = VecZeroEntries(DDD);CHKERRQ(ierr); 
        ierr = MatDiagonalSet(NSymmA,DDD,INSERT_VALUES);CHKERRQ(ierr);
        
        /*D = sum(abs(NSymmA'));*/
        
        ierr = absmat(NSymmA,&NSymmAtrabs);CHKERRQ(ierr);                       
        ierr = MatCreateVecs(NSymmAtrabs,&D,NULL);CHKERRQ(ierr);    
        ierr = VecZeroEntries(D);CHKERRQ(ierr);
        ierr = MatAssemblyBegin(NSymmAtrabs,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
        ierr = MatAssemblyEnd(NSymmAtrabs,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
        ierr = MatGetRowSum(NSymmAtrabs,D);CHKERRQ(ierr);
        
        /*L = diag(D) + NSymmA;*/
        
        ierr = MatDuplicate(NSymmA,MAT_COPY_VALUES,&L);CHKERRQ(ierr);
        ierr = MatSetOption(L,MAT_NEW_NONZERO_ALLOCATION_ERR,PETSC_FALSE);CHKERRQ(ierr); 
        ierr = MatDiagonalSet(L,D,ADD_VALUES);CHKERRQ(ierr);

    } else if (wgh) { /*weighted Laplacian*/
        
        /* initialize D as zero vector */
        
        ierr = MatCreateVecs(SymmA,&D,NULL);CHKERRQ(ierr);
        ierr = VecZeroEntries(D);CHKERRQ(ierr);
        ierr = VecDuplicate(D,&DD);;CHKERRQ(ierr);

        /* D = sum(SymmA); */ 
        ierr = MatAssemblyBegin(SymmA,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
        ierr = MatAssemblyEnd(SymmA,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
        ierr = MatGetRowSum(SymmA,D);CHKERRQ(ierr);
        
        /*L = -(SymmA-diag(diag(SymmA)))+diag(D); 
        Bu ayni zamanda L = -SymmA (diagonal sifir) +diag(D); demek
        Iki satirda: L = -SymmA;
                    L = L+diag(D); olarak yapildi*/
        
        ierr = MatDuplicate(SymmA,MAT_DO_NOT_COPY_VALUES,&L);CHKERRQ(ierr);
        ierr = MatSetOption(SymmA,MAT_NEW_NONZERO_ALLOCATION_ERR,PETSC_FALSE);CHKERRQ(ierr);
        ierr = MatDiagonalSet(SymmA,DD,INSERT_VALUES);CHKERRQ(ierr);
        ierr = MatAXPY(L,-1.,SymmA,DIFFERENT_NONZERO_PATTERN);CHKERRQ(ierr);
        ierr = MatSetOption(L,MAT_NEW_NONZERO_ALLOCATION_ERR,PETSC_FALSE);CHKERRQ(ierr);
        ierr = MatDiagonalSet(L,D,ADD_VALUES);CHKERRQ(ierr);

    } 

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
                Create the eigensolver and set various options
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  /*
     Create eigensolver context
  */
  ierr = EPSCreate(PETSC_COMM_WORLD,&eps);CHKERRQ(ierr);

  /*
     Set operators. In this case, it is a standard eigenvalue problem
  */
  ierr = EPSSetOperators(eps,L,NULL);CHKERRQ(ierr);        
  ierr = EPSSetProblemType(eps,EPS_HEP);CHKERRQ(ierr);

  /*
     Select portion of spectrum
  */
  ierr = EPSSetWhichEigenpairs(eps,EPS_SMALLEST_MAGNITUDE);CHKERRQ(ierr);         

  /*
     Set solver parameters at runtime
  */
  ierr = EPSSetFromOptions(eps);CHKERRQ(ierr);

  /*
     Attach deflation space: in this case, the matrix has a constant
     nullspace, [1 1 ... 1]^T is the eigenvector of the zero eigenvalue
  */
  ierr = MatCreateVecs(L,&x,NULL);CHKERRQ(ierr);       
  ierr = VecSet(x,1.0);CHKERRQ(ierr);
  ierr = EPSSetDeflationSpace(eps,1,&x);CHKERRQ(ierr);
  ierr = VecDestroy(&x);CHKERRQ(ierr);

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
                      Solve the eigensystem
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  ierr = EPSSolve(eps);CHKERRQ(ierr);

  /*
     Optional: Get some information from the solver and display it
  */
  ierr = EPSGetType(eps,&type);CHKERRQ(ierr);
  ierr = PetscPrintf(PETSC_COMM_WORLD," Solution method: %s\n\n",type);CHKERRQ(ierr);
  ierr = EPSGetDimensions(eps,&nev,NULL,NULL);CHKERRQ(ierr);
  ierr = PetscPrintf(PETSC_COMM_WORLD," Number of requested eigenvalues: %D\n",nev);CHKERRQ(ierr);
  
  ierr = MatCreateVecs(L,&vr,NULL);CHKERRQ(ierr);
  ierr = EPSGetEigenpair(eps,0,&kr,NULL,vr,NULL);  
  ierr = PetscPrintf(PETSC_COMM_WORLD," The second smallest eigenvalue: %g\n",kr);CHKERRQ(ierr);    
  
  /* sort second smallest eigenvector */

  ierr = VecGetSize(vr,&siz);CHKERRQ(ierr);
  ierr = PetscMalloc1(siz,&idx);CHKERRQ(ierr);
  for (i=0; i<siz;i++) idx[i] = i;
  
  VecScatter ctx;
  Vec vout;
  VecScatterCreateToAll(vr,&ctx,&vout);
  VecScatterBegin(ctx,vr,vout,INSERT_VALUES,SCATTER_FORWARD);
  VecScatterEnd(ctx,vr,vout,INSERT_VALUES,SCATTER_FORWARD);
  //VecScatterDestroy(ctx);
  
  PetscScalar *avr;
  ierr = VecGetArray(vout,&avr);CHKERRQ(ierr);
  ierr = PetscSortRealWithPermutation(siz,avr,idx);CHKERRQ(ierr);    

  PetscMPIInt rank,size;
  MPI_Comm_rank(PETSC_COMM_WORLD, &rank);
  MPI_Comm_size(PETSC_COMM_WORLD, &size);
 

  PetscInt mod;
  mod = siz % size;
  
  PetscInt *idxx,ss;

  ss = (siz-mod)/size;                                                   

  PetscMalloc1(ss,&idxx);
  if (rank != size-1) {
    j =0;
    for (i=rank*ss; i<(rank+1)*ss; i++) {
      idxx[j] = idx[i];
      j++;
    }

  } else {
      
    j =0;
    for (i=rank*ss; i<siz; i++) {
      idxx[j] = idx[i];
      j++;
    }
      
  }
  

  if (mod != 0){
    if (rank<mod){
        PetscMalloc1(ss+1,&idxx);                                       
        idxx[ss+1] = idx[ss*size+rank+1];
    }
  }
 
  /*Permute matrix L (spy(A(p1,p1))*/
  ierr = ISCreateGeneral(PETSC_COMM_WORLD,ss,idxx,PETSC_COPY_VALUES,&is);CHKERRQ(ierr);  
  ierr = ISSetPermutation(is);CHKERRQ(ierr);
//ISView(is,PETSC_VIEWER_STDOUT_WORLD);

  ierr = MatPermute(A,is,is,&PL);CHKERRQ(ierr);  
    /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
                    Create Partitioning
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  
  ierr = MatConvert(PL,MATMPIADJ,MAT_INITIAL_MATRIX,&AL);CHKERRQ(ierr);                    

  ierr = MatPartitioningCreate(MPI_COMM_WORLD,&part);CHKERRQ(ierr);
  ierr = MatPartitioningSetAdjacency(part,AL);CHKERRQ(ierr);                             
  ierr = MatPartitioningSetFromOptions(part);CHKERRQ(ierr);
  ierr = MatPartitioningApply(part,&partitioning);CHKERRQ(ierr);
  //ierr = ISView(partitioning,PETSC_VIEWER_STDOUT_WORLD);CHKERRQ(ierr);          
  MatPartitioningView(part,	PETSC_VIEWER_STDOUT_WORLD);
  ierr = ISDestroy(&partitioning);CHKERRQ(ierr);
  ierr = MatPartitioningDestroy(&part);CHKERRQ(ierr);

  ierr = VecDestroy(&DD);CHKERRQ(ierr);
  ierr = VecDestroy(&D);CHKERRQ(ierr);
  ierr = MatDestroy(&AL);CHKERRQ(ierr);
  ierr = MatDestroy(&PL);CHKERRQ(ierr);
  ierr = MatDestroy(&SymmA);CHKERRQ(ierr);
  ierr = MatDestroy(&Atr);CHKERRQ(ierr); 
  ierr = MatDestroy(&A);CHKERRQ(ierr);                      
  ierr = MatDestroy(&L);CHKERRQ(ierr); 
  ierr = VecDestroy(&vr);CHKERRQ(ierr);
  ierr = EPSDestroy(&eps);CHKERRQ(ierr);

  ierr = SlepcFinalize();
  return ierr;
} 


/*TEST

   build:
      requires:  !complex !define(PETSC_USE_64BIT_INDICES), parmetis

   test:
      size: 2
      args: -f ${wPETSC_DIR}/share/petsc/datafiles/matrices/binary_files/airfoil1_binary -unweighted -mat_partitioning_type parmetis
      
   usage: mpiexec -n 2 ./example -f /home/edaoktay/petsc-3.10.3/share/petsc/datafiles/matrices/binary_files/airfoil1_binary -unweighted -mat_partitioning_type parmetis

TEST*/
